{% extends "_base.html" %}
{% block title %}Your email has now been validated{% endblock %}

{% block body %}
<p>Your email has now been validated, please return to your client. You may now close this window.</p>
{% endblock %}
