{% extends "_base.html" %}
{% block title %}SSO account deactivated{% endblock %}

{% block header %}
<style type="text/css">
    {% include "sso.css" without context %}
</style>
{% endblock %}

{% block body %}
<div class="error_page">
    <header>
        <h1>Your account has been deactivated</h1>
        <p>
            <strong>No account found</strong>
        </p>
        <p>
            Your account might have been deactivated by the server administrator.
            You can either try to create a new account or contact the server’s
            administrator.
        </p>
    </header>
</div>
{% include "sso_footer.html" without context %}
{% endblock %}
