{% extends "_base.html" %}
{% block title %}Authentication successful{% endblock %}

{% block header %}
<style type="text/css">
    {% include "sso.css" without context %}
</style>
<script>
    if (window.onAuthDone) {
        window.onAuthDone();
    } else if (window.opener && window.opener.postMessage) {
        window.opener.postMessage("authDone", "*");
    }
</script>
{% endblock %}

{% block body %}
<header>
    <h1>Thank you</h1>
    <p>
        Now we know it’s you, you can close this window and return to the
        application.
    </p>
</header>
{% include "sso_footer.html" without context %}
{% endblock %}
